# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

import sys
from collections import namedtuple
from enum import Enum
from typing import (
    Any,
    Callable,
    Optional,
    Type,
    TypeVar,
    Union,
)

import numpy as np
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.preconditions import check_argument

if sys.version_info >= (3, 10):
    from typing import (
        Concatenate,
        ParamSpec,
    )
else:
    from typing_extensions import (
        Concatenate,
        ParamSpec,
    )


T = TypeVar("T", bound=Enum)


def validate_enum(enum_: Type[T], item: T | str) -> str:
    """
    Check whether the item is a valid option in enum_. If so, return the name of that option.
    Otherwise, raise an error.

    Parameters
    ----------
    enum_ : T
        An Enum where we expect the option value to be a str.
    item : T or str
        The item to be checked with enum_.

    Returns
    -------
    str
        The name of a valid enum option.
    """
    if isinstance(item, enum_):
        return item.name
    try:
        return getattr(enum_, item).name  # type: ignore
    except (TypeError, AttributeError) as err:
        raise QctrlArgumentsValueError(
            "Only the following options are allowed: " f"{list(enum_.__members__)} ",
            arguments={"item": item},
        ) from err


def _is_integer(val: Any) -> bool:
    return isinstance(val, ScalarDType.INT.value.types)


def _is_real(val: Any) -> bool:
    return _is_integer(val) or isinstance(val, ScalarDType.REAL.value.types)


def _is_complex(val: Any) -> bool:
    return _is_real(val) or isinstance(val, ScalarDType.COMPLEX.value.types)


def _is_number(val: Any) -> bool:
    return _is_real(val) or _is_complex(val)


def _number_converter(val: Any) -> int | float | complex:
    if _is_integer(val):
        return int(val)
    if _is_real(val):
        return float(val)
    return complex(val)


# types: supported types defined by Python and Numpy for a given dtype
# checker: a callable to check the input scalar
# converter: a callable to convert the scalar to the corresponding Python primitive type
_ScalarDTypeValidator = namedtuple(
    "_ScalarDTypeValidator", ["types", "checker", "converter"]
)

_SCALAR = TypeVar(
    "_SCALAR", bound=Union[int, float, complex, np.integer, np.float_, np.complex_]
)


class ScalarDType(Enum):
    """
    Store dtypes to validate both Python and NumPy types.
    """

    INT = _ScalarDTypeValidator((int, np.integer), _is_integer, int)
    REAL = _ScalarDTypeValidator((float, np.float_), _is_real, float)
    COMPLEX = _ScalarDTypeValidator((complex, np.complex_), _is_complex, complex)
    NUMBER = _ScalarDTypeValidator(None, _is_number, _number_converter)

    def __call__(
        self,
        value: _SCALAR,
        name: str,
        min_: Optional[Any] = None,
        max_: Optional[Any] = None,
        min_inclusive: bool = False,
        max_inclusive: bool = False,
    ) -> _SCALAR:
        """
        Validate a given scalar by the dtype.
        If valid, return the value as a primitive Python numeric type.
        """
        check_argument(
            self.value.checker(value),
            f"The {name} must be a {self.name.lower()}.",
            {name: value},
        )

        if min_ is not None:
            check_argument(
                value > min_ or (min_inclusive and value == min_),
                f"The {name} must be greater than {'or equal to' if min_inclusive  else ''} "
                f"{min_}.",
                {name: value},
            )
        if max_ is not None:
            check_argument(
                value < max_ or (max_inclusive and value == max_),
                f"The {name} must be smaller than {'or equal to' if max_inclusive  else ''} "
                f"{max_}.",
                {name: value},
            )
        return self.value.converter(value)


# valid_dtype_kinds: valid values for the array's dtype.kind.
# dtype: data type of the returned validated NumPy array
_ArrayDTypeValidator = namedtuple(
    "_ArrayDTypeValidator", ["valid_dtype_kinds", "dtype"]
)


class ArrayDType(Enum):
    """
    Store dtypes to validate array-likes.
    """

    INT = _ArrayDTypeValidator("iu", np.integer)
    REAL = _ArrayDTypeValidator("iuf", np.float_)
    COMPLEX = _ArrayDTypeValidator("iufc", np.complex_)
    NUMERIC = _ArrayDTypeValidator("iufc", None)

    def __call__(
        self,
        value: Any,
        name: str,
        ndim: Optional[int] = None,
        shape: Optional[tuple] = None,
        min_: Optional[Any] = None,
        max_: Optional[Any] = None,
        min_inclusive: bool = False,
        max_inclusive: bool = False,
    ) -> np.ndarray:
        """
        Validate a given array-like by the dtype.
        If valid, return the value as a NumPy array with the corresponding dtype.
        """
        array_val = np.asarray(value)

        check_argument(
            array_val.size > 0, f"The {name} must not be an empty array.", {name: value}
        )

        check_argument(
            array_val.dtype.kind in self.value.valid_dtype_kinds,
            f"The {name} must be a {self.name.lower()} array.",
            {name: value},
        )

        if ndim is not None:
            check_argument(
                array_val.ndim == ndim,
                f"The {name} must be a {ndim}D array.",
                {name: value},
            )
        if shape is not None:
            check_argument(
                array_val.shape == shape,
                f"The {name} must be an array with shape {shape}.",
                {name: value},
            )

        if min_ is not None:
            check_argument(
                np.all(array_val > min_)
                or (min_inclusive and np.all(array_val >= min_)),
                f"The values in the {name} must be greater than "
                f"{'or equal to' if min_inclusive  else ''} {min_}.",
                {name: value},
            )
        if max_ is not None:
            check_argument(
                np.all(array_val < max_)
                or (max_inclusive and np.all(array_val <= max_)),
                f"The values in the {name} must be smaller than "
                f"{'or equal to' if max_inclusive  else ''} {max_}.",
                {name: value},
            )
        # Special handler for cases:
        # - int typed array to address the signed and unsigned cases.
        # - a generic numeric array
        # That is, we preserve the dtype from users in these cases.
        if self in (ArrayDType.INT, ArrayDType.NUMERIC):
            return array_val
        try:
            return array_val.astype(self.value.dtype, casting="safe")
        except TypeError as err:
            raise QctrlArgumentsValueError(
                f"Expected {name} as an array of {self.name.lower()} dtype, "
                f"but got {array_val.dtype}.",
                {name: value},
            ) from err


_VAL = TypeVar("_VAL")
P = ParamSpec("P")


def nullable(
    validator: Callable[Concatenate[_VAL, P], _VAL],
    value: Optional[_VAL],
    *args: P.args,
    **kwargs: P.kwargs,
) -> Optional[_VAL]:
    """
    Validate a parameter that can be None.

    When the parameter holds a non-null value, the validator callable is used to check the value.
    The validator takes the value as the first argument and some other options as defined by P, it
    returns the same type as the input value (strictly speaking, the returned type is something
    that can be converted from the input one. But in reality, we expect  we expect them to be
    interchangeable). The P annotation here allows mypy to also check the types for the
    resting arguments of the validator.
    """
    if value is None:
        return value
    return validator(value, *args, **kwargs)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from __future__ import annotations

from typing import ContextManager

from qctrlcommons.exceptions import QctrlException
from qctrlcommons.preconditions import check_argument
from qctrlworkflowclient import (
    ApiRouter,
    LocalRouter,
    get_authenticated_client_for_product,
)
from qctrlworkflowclient.router.api import Action
from rich.console import Console
from rich.table import Table

from boulderopal._configuration.configuration import (
    configure,
    get_configuration,
)
from boulderopal._constants import INVALID_SUBSCRIPTION_ERROR
from boulderopal._core.formatter import (
    async_result_formatter,
    metadata_formatter,
)
from boulderopal._validation.basic import ScalarDType


def get_cloud_router(on_fail_message: str) -> ApiRouter:
    """
    Return the current cloud router.

    Parameters
    ----------
    on_fail_message : str
        Message to throw when running locally.

    Returns
    -------
    ApiRouter
        The current ApiRouter object.

    Raises
    ------
    QctrlException
        If running in local mode.
    """
    router = get_configuration().get_router()
    if isinstance(router, LocalRouter):
        raise QctrlException(f"{on_fail_message} is not supported in local mode.")
    return router


def get_result(action_id: int | str) -> dict:
    """
    Retrieve the result of a previously run calculation.

    Parameters
    ----------
    action_id : int or str
        The ID associated to the calculation.
        You can get the ID of your calculations from the activity monitor.

    Returns
    -------
    dict
        A dictionary containing the calculation result.
    """
    check_argument(
        str(action_id).isnumeric(),
        "The calculation id must be an integer.",
        {"action_id": action_id},
    )
    action = Action(action_id=str(action_id))
    return metadata_formatter(
        get_cloud_router("Retrieving results").get_result(action)
    ).decoded


def request_machines(machine_count: int) -> None:
    """
    Request a minimum number of machines to be online.

    Parameters
    ----------
    machine_count : int
        The minimum number of machines requested to be online.

    Notes
    -----
    This command is blocking until the specified amount of machines
    have been observed in your environment. It only attempts to ensure
    the requested amount are online, not necessarily starting up that
    amount if some machines are already online.
    """
    machine_count = ScalarDType.INT(
        machine_count, "machine_count", min_=1, min_inclusive=True
    )

    get_cloud_router("Requesting machines").request_machines(machine_count)


def group_requests() -> ContextManager:
    """
    Create a context manager for executing multiple function calls over available machines.

    Returns
    -------
    ContextManager
        A context manager to collect and run computation requests.

    Notes
    -----
    You can to group up to five requests together.
    All grouped calculations must be independent from each other.

    Within the context manager, the object returned from each request is a placeholder.
    When exiting, the context manager waits until all calculations have finished,
    hence this command blocks execution.
    When all results are received, the placeholders are replaced with them.

    Read the `Computational resources in Boulder Opal
    <https://docs.q-ctrl.com/boulder-opal/topics/computational-resources-in-boulder-opal>`_
    topic for more information about this feature.
    """
    return get_cloud_router("Grouping requests").enable_parallel(
        callback=async_result_formatter
    )


def set_organization(organization_slug: str) -> None:
    """
    Set the organization for Boulder Opal cloud calls.

    Parameters
    ----------
    organization_slug : str
        Unique slug for the organization.
    """
    configure(organization=organization_slug)


def set_api(api_url: str, oidc_url: str) -> None:
    """
    Configure Boulder Opal for API routing.

    Parameters
    ----------
    api_url : str
        URL of the GraphQL schema.
    oidc_url : str
        Base URL of the OIDC provider, e.g. Keycloak.
    """
    client = get_authenticated_client_for_product(
        product_access_required="boulder-opal-cli-access",
        api_url=api_url,
        auth=oidc_url,
        invalid_access_error_message=INVALID_SUBSCRIPTION_ERROR,
    )
    settings = get_configuration()

    configure(router=ApiRouter(client, settings))


def show_machine_status() -> None:
    """
    Print the current machine status in Boulder Opal cloud environment for your organization.
    """

    status_map = {
        "Online": "online",
        "Starting up": "initializing",
        "Pending": "pending",
        "Shutting down": "terminating",
        "Offline": "offline",
    }

    status = get_cloud_router("Machine status").get_machine_status()

    table = Table(title="Machine status", show_lines=True)
    console = Console()

    for header in ["Status", "Count"]:
        table.add_column(header, justify="center")

    for state, key in status_map.items():
        number = status.get(key)
        if number is None:
            raise RuntimeError(
                f"Unknown machine state. Expected {list(status.keys())}, got {key}."
            )
        table.add_row(state, str(number))

    console.print(table)



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

# pylint: disable=missing-function-docstring
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
)

import numpy as np

from .base import boulder_opal_workflow

if TYPE_CHECKING:
    from boulderopal.graph import Graph


@boulder_opal_workflow("closed_loop_step_workflow")
def closed_loop_step_workflow(
    optimizer: dict, results: Optional[dict], test_point_count: Optional[int]
) -> dict[str, Any]:
    return {
        "optimizer": optimizer,
        "results": results,
        "test_point_count": test_point_count,
    }


@boulder_opal_workflow("execute_graph_workflow")
def execute_graph_workflow(
    graph: "Graph", output_node_names: list[str], execution_mode: str
) -> dict[str, Any]:
    return {
        "graph": graph,
        "output_node_names": output_node_names,
        "execution_mode": execution_mode,
    }


@boulder_opal_workflow("obtain_ion_chain_properties_workflow")
def obtain_ion_chain_properties_workflow(
    atomic_mass: float,
    ion_count: int,
    center_of_mass_frequencies: np.ndarray,
    wavevector: np.ndarray,
    laser_detuning: Optional[float],
) -> dict[str, Any]:
    return {
        "atomic_mass": atomic_mass,
        "ion_count": ion_count,
        "center_of_mass_frequencies": center_of_mass_frequencies,
        "wavevector": wavevector,
        "laser_detuning": laser_detuning,
    }


@boulder_opal_workflow("reconstruct_noise_workflow")
def reconstruct_noise_workflow(
    method: dict[str, dict[str, Any]],
    noises_frequencies: list[np.ndarray],
    filter_functions: list[np.ndarray],
    infidelities: np.ndarray,
    infidelity_uncertainties: Optional[np.ndarray],
) -> dict[str, Any]:
    return {
        "method": method,
        "noises_frequencies": noises_frequencies,
        "filter_functions": filter_functions,
        "infidelities": infidelities,
        "infidelity_uncertainties": infidelity_uncertainties,
    }


@boulder_opal_workflow("run_gradient_free_optimization_workflow")
def run_gradient_free_optimization_workflow(
    graph: "Graph",
    cost_node_name: str,
    output_node_names: list[str],
    iteration_count: int,
    target_cost: Optional[float],
    optimization_count: int,
    cost_history_scope: str,
    seed: Optional[int],
) -> dict[str, Any]:
    return {
        "graph": graph,
        "cost_node_name": cost_node_name,
        "output_node_names": output_node_names,
        "iteration_count": iteration_count,
        "target_cost": target_cost,
        "optimization_count": optimization_count,
        "cost_history_scope": cost_history_scope,
        "seed": seed,
    }


@boulder_opal_workflow("run_optimization_workflow")
def run_optimization_workflow(
    graph: "Graph",
    optimization_count: int,
    cost_node_name: str,
    output_node_names: list[str],
    target_cost: Optional[float],
    max_iteration_count: Optional[int],
    cost_tolerance: Optional[float],
    cost_history_scope: str,
    seed: Optional[int],
) -> dict[str, Any]:
    return {
        "graph": graph,
        "optimization_count": optimization_count,
        "cost_node_name": cost_node_name,
        "output_node_names": output_node_names,
        "target_cost": target_cost,
        "max_iteration_count": max_iteration_count,
        "cost_tolerance": cost_tolerance,
        "cost_history_scope": cost_history_scope,
        "seed": seed,
    }


@boulder_opal_workflow("run_stochastic_optimization_workflow")
def run_stochastic_optimization_workflow(
    graph: "Graph",
    optimizer: dict,
    cost_node_name: str,
    output_node_names: list[str],
    iteration_count: int,
    cost_history_scope: str,
    target_cost: Optional[float],
    seed: Optional[int],
) -> dict[str, Any]:
    return {
        "graph": graph,
        "optimizer": optimizer,
        "cost_node_name": cost_node_name,
        "output_node_names": output_node_names,
        "iteration_count": iteration_count,
        "cost_history_scope": cost_history_scope,
        "target_cost": target_cost,
        "seed": seed,
    }



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.
from __future__ import annotations

from typing import Optional

import numpy as np
from qctrlcommons.preconditions import check_argument

from boulderopal._core import obtain_ion_chain_properties_workflow
from boulderopal._validation.basic import (
    ArrayDType,
    ScalarDType,
)


def obtain_ion_chain_properties(
    atomic_mass: float,
    ion_count: int,
    center_of_mass_frequencies: np.ndarray,
    wavevector: np.ndarray,
    laser_detuning: Optional[float] = None,
) -> dict:
    """
    Calculate the Lamb–Dicke parameters, frequencies (or relative detunings
    if a laser detuning is provided), and eigenvectors
    of the collective motional modes of an ion chain.

    Parameters
    ----------
    atomic_mass : float
        The atomic mass of the ions of the chain in atomic units.
        All ions in the chain are assumed to be of the same species.
    ion_count : int
        The number of ions in the chain, :math:`N`.
    center_of_mass_frequencies : np.ndarray
        The center-of-mass trapping frequencies in each direction.
        Must contain three positive elements.
    wavevector : np.ndarray
        The laser difference angular wave vector (in rad/m) in each direction.
        Must contain three elements.
    laser_detuning : float or None, optional
        The detuning of the control laser.
        If not provided, the returned relative detunings represent the mode frequencies.

    Returns
    -------
    dict
        A dictionary containing the ion chain properties, with the following keys:

        ``lamb_dicke_parameters``
            A 3D array of shape ``(3, N, N)`` representing the Lamb–Dicke parameters of the ions.
            Its dimensions indicate, respectively, direction, mode, and ion.
        ``relative_detunings``
            A 2D array of shape ``(3, N)`` representing the mode frequencies
            (or relative detunings if a laser detuning is provided).
            Its dimensions indicate, respectively, direction and mode.
        ``eigenvectors``
            A 3D array of shape ``(3, N, N)`` representing the eigenvectors of each mode.
            Its dimensions indicate, respectively, direction, mode, and ion.
        ``metadata``
            Metadata associated with the calculation.
            No guarantees are made about the contents of this metadata dictionary;
            the contained information is intended purely to help interpret the results of the
            calculation on a one-off basis.

    See Also
    --------
    boulderopal.ions.ms_optimize :
        Find optimal pulses to perform Mølmer–Sørensen-type operations on trapped ions systems.
    boulderopal.ions.ms_simulate :
        Simulate a Mølmer–Sørensen-type operation on a trapped ions system.

    Notes
    -----
    The directions of input parameters and returned arrays are ordered as
    radial x-direction, radial y-direction, and axial z-direction, corresponding, respectively,
    to the unit vectors :math:`(1, 0, 0)`, :math:`(0, 1, 0)`, and :math:`(0, 0, 1)`.

    Examples
    --------
    Refer to the `How to optimize error-robust Mølmer–Sørensen gates for trapped ions
    <https://docs.q-ctrl.com/boulder-opal/user-guides/how-to-optimize-error-robust
    -molmer-sorensen-gates-for-trapped-ions>`_ user guide to find how to use this function.
    """

    center_of_mass_frequencies = ArrayDType.REAL(
        center_of_mass_frequencies,
        "center_of_mass_frequencies",
        ndim=1,
        shape=(3,),
        min_=0,
    )
    wavevector = ArrayDType.REAL(wavevector, "wavevector", ndim=1, shape=(3,))
    check_argument(
        not np.allclose(wavevector, 0),
        "At least one of the wavevector components must be non-zero.",
        {"wavevector": wavevector},
    )

    return obtain_ion_chain_properties_workflow(
        atomic_mass=ScalarDType.REAL(atomic_mass, "atomic_mass", min_=0),
        ion_count=ScalarDType.INT(ion_count, "ion_count", min_=0),
        center_of_mass_frequencies=center_of_mass_frequencies,
        wavevector=wavevector,
        laser_detuning=laser_detuning,
    )



# Copyright 2023 Q-CTRL. All rights reserved.
#
# Licensed under the Q-CTRL Terms of service (the "License"). Unauthorized
# copying or use of this file, via any medium, is strictly prohibited.
# Proprietary and confidential. You may not use this file except in compliance
# with the License. You may obtain a copy of the License at
#
#    https://q-ctrl.com/terms
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS. See the
# License for the specific language.

from typing import (
    List,
    Union,
)

import forge
from qctrlcommons.exceptions import QctrlArgumentsValueError
from qctrlcommons.node.base import Node
from qctrlcommons.node.wrapper import NamedNodeData

from .documentation import Category
from .node_data import (
    FilterFunction,
    Pwc,
)


class GetAttributeNode(Node):
    """
    Get an attribute from a node value.

    Typically you would use the syntax ``value.attr`` instead of using this function directly.

    Parameters
    ----------
    value : Any
        The value from which to get the item.
    attr : str
        The name of the attribute.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Any
        The item or items obtained from ``value.attr``, or to be more precise,
        ``getattr(value, attr)``.

    Notes
    -----
    Only certain combinations of `value` and `attr` are supported:

    ``Pwc.values``
        The tensor of values of a `Pwc`.
    """

    name = "getattr"
    args = [forge.arg("value"), forge.arg("attr", type=str)]
    categories = [Category.OTHER_OPERATIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        value = kwargs.get("value")
        attr = kwargs.get("attr")

        valid_combinations = [
            (FilterFunction, "inverse_powers"),
            (FilterFunction, "uncertainties"),
            (Pwc, "values"),
        ]
        for valid_value_type, valid_attr in valid_combinations:
            if isinstance(value, valid_value_type) and attr == valid_attr:
                return NamedNodeData(_operation)

        # Invalid combination of value type and attr, so show an error message.
        raise QctrlArgumentsValueError(
            "Not allowed to fetch attribute",
            kwargs,
            extras={
                "Valid combinations": [
                    f"{valid_value_type.__name__}.{valid_attr}"
                    for valid_value_type, valid_attr in valid_combinations
                ]
            },
        )


class GetItemNode(Node):
    """
    Get an item (or items) from a node value.

    Typically you would use slicing syntax ``value[key]`` instead of
    using this function directly.

    Parameters
    ----------
    value : Any
        The value from which to get the item.
    key : int or slice or list[int or slice]
        The key for the item or items.
    name : str or None, optional
        The name of the node.

    Returns
    -------
    Any
        The item or items obtained from ``value[key]``.
    """

    name = "getitem"
    args = [
        forge.arg("value"),
        forge.arg("key", type=Union[int, slice, List[Union[int, slice]]]),
    ]
    categories = [Category.OTHER_OPERATIONS]

    @classmethod
    def create_node_data(cls, _operation, **kwargs):
        return NamedNodeData(_operation)
